# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import itertools
from unittest import mock

import pytest  # type: ignore

import google_crc32c

EMPTY = b""
EMPTY_CRC = 0x00000000

# From: https://tools.ietf.org/html/rfc3720#appendix-B.4
#
#   32 bytes of zeroes:
#
#     Byte:        0  1  2  3
#
#        0:       00 00 00 00
#      ...
#       28:       00 00 00 00
#
#      CRC:       aa 36 91 8a

ALL_ZEROS = b"\x00" * 32
ALL_ZEROS_CRC = 0x8A9136AA

#   32 bytes of ones:
#
#     Byte:        0  1  2  3
#
#        0:       ff ff ff ff
#      ...
#       28:       ff ff ff ff
#
#      CRC:       43 ab a8 62

ALL_ONES = b"\xff" * 32
ALL_ONES_CRC = 0x62A8AB43

#
#   32 bytes of incrementing 00..1f:
#
#     Byte:        0  1  2  3
#
#        0:       00 01 02 03
#      ...
#       28:       1c 1d 1e 1f
#
#      CRC:       4e 79 dd 46

INCREMENTING = bytes(range(32))
INCREMENTING_CRC = 0x46DD794E

#
#   32 bytes of decrementing 1f..00:
#
#     Byte:        0  1  2  3
#
#        0:       1f 1e 1d 1c
#      ...
#       28:       03 02 01 00
#
#      CRC:       5c db 3f 11

DECREMENTING = bytes(reversed(range(32)))
DECREMENTING_CRC = 0x113FDB5C

#
#    An iSCSI - SCSI Read (10) Command PDU
#
#    Byte:        0  1  2  3
#
#       0:       01 c0 00 00
#       4:       00 00 00 00
#       8:       00 00 00 00
#      12:       00 00 00 00
#      16:       14 00 00 00
#      20:       00 00 04 00
#      24:       00 00 00 14
#      28:       00 00 00 18
#      32:       28 00 00 00
#      36:       00 00 00 00
#      40:       02 00 00 00
#      44:       00 00 00 00
#
#     CRC:       56 3a 96 d9

ISCSI_SCSI_READ_10_COMMAND_PDU = [
    0x01,
    0xC0,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x14,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x04,
    0x00,
    0x00,
    0x00,
    0x00,
    0x14,
    0x00,
    0x00,
    0x00,
    0x18,
    0x28,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x02,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
    0x00,
]
ISCSI_LENGTH = len(ISCSI_SCSI_READ_10_COMMAND_PDU)
ISCSI_BYTES = bytes(ISCSI_SCSI_READ_10_COMMAND_PDU)
ISCSI_CRC = 0xD9963A56


def iscsi_chunks(chunksize):
    for index in itertools.islice(range(ISCSI_LENGTH), 0, None, chunksize):
        yield ISCSI_SCSI_READ_10_COMMAND_PDU[index : index + chunksize]


_EXPECTED = [
    (EMPTY, EMPTY_CRC),
    (ALL_ZEROS, ALL_ZEROS_CRC),
    (ALL_ONES, ALL_ONES_CRC),
    (INCREMENTING, INCREMENTING_CRC),
    (DECREMENTING, DECREMENTING_CRC),
    (ISCSI_SCSI_READ_10_COMMAND_PDU, ISCSI_CRC),
    (ISCSI_BYTES, ISCSI_CRC),
]


def test_extend_w_empty_chunk():
    crc = 123
    assert google_crc32c.extend(crc, b"") == crc


def test_extend_w_multiple_chunks():
    crc = 0

    for chunk in iscsi_chunks(7):
        chunk_bytes = bytes(chunk)
        crc = google_crc32c.extend(crc, chunk_bytes)

    assert crc == ISCSI_CRC


def test_extend_w_reduce():
    chunks = [bytes(chunk) for chunk in iscsi_chunks(3)]
    assert functools.reduce(google_crc32c.extend, chunks, 0) == ISCSI_CRC


@pytest.mark.parametrize("chunk, expected", _EXPECTED)
def test_value(chunk, expected):
    assert google_crc32c.value(bytes(chunk)) == expected


def pytest_generate_tests(metafunc):
    if "_crc32c" in metafunc.fixturenames:
        metafunc.parametrize("_crc32c", ["python", "cext"], indirect=True)

@pytest.fixture
def _crc32c(request):
    if request.param == "python":
        from google_crc32c import python
        return python
    elif request.param == "cext":
        from google_crc32c import cext
        return cext
    else:
        raise ValueError("invalid internal test config")


class TestChecksum(object):
    @staticmethod
    def test_ctor_defaults(_crc32c):
        helper = google_crc32c.Checksum()
        assert helper._crc == 0

    @staticmethod
    def test_ctor_explicit(_crc32c):
        chunk = b"DEADBEEF"
        helper = google_crc32c.Checksum(chunk)
        assert helper._crc == google_crc32c.value(chunk)

    @staticmethod
    def test_update(_crc32c):
        chunk = b"DEADBEEF"
        helper = google_crc32c.Checksum()
        helper.update(chunk)
        assert helper._crc == google_crc32c.value(chunk)

    @staticmethod
    def test_update_w_multiple_chunks(_crc32c):
        helper = google_crc32c.Checksum()

        for index in itertools.islice(range(ISCSI_LENGTH), 0, None, 7):
            chunk = ISCSI_SCSI_READ_10_COMMAND_PDU[index : index + 7]
            helper.update(bytes(chunk))

        assert helper._crc == ISCSI_CRC

    @staticmethod
    def test_digest_zero(_crc32c):
        helper = google_crc32c.Checksum()
        assert helper.digest() == b"\x00" * 4

    @staticmethod
    def test_digest_nonzero(_crc32c):
        helper = google_crc32c.Checksum()
        helper._crc = 0x01020304
        assert helper.digest() == b"\x01\x02\x03\x04"

    @staticmethod
    def test_hexdigest_zero(_crc32c):
        helper = google_crc32c.Checksum()
        assert helper.hexdigest() == b"00" * 4

    @staticmethod
    def test_hexdigest_nonzero(_crc32c):
        helper = google_crc32c.Checksum()
        helper._crc = 0x091A3B2C
        assert helper.hexdigest() == b"091a3b2c"

    @staticmethod
    def test_copy(_crc32c):
        chunk = b"DEADBEEF"
        helper = google_crc32c.Checksum(chunk)
        clone = helper.copy()
        before = helper._crc
        helper.update(b"FACEDACE")
        assert clone._crc == before

    @staticmethod
    @pytest.mark.parametrize("chunksize", [1, 3, 5, 7, 11, 13, ISCSI_LENGTH])
    def test_consume_stream(_crc32c, chunksize):
        helper = google_crc32c.Checksum()
        expected = [bytes(chunk) for chunk in iscsi_chunks(chunksize)]
        stream = mock.Mock(spec=["read"])
        stream.read.side_effect = expected + [b""]

        found = list(helper.consume(stream, chunksize))

        assert helper._crc == ISCSI_CRC
        assert found == expected
        for call in stream.read.call_args_list:
            assert call == mock.call(chunksize)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

_SLOW_CRC32C_WARNING = (
    "As the c extension couldn't be imported, `google-crc32c` is using a " 
    "pure python implementation that is significantly slower. If possible, "
    "please configure a c build environment and compile the extension"
)

# Default to C exstension Implementation, falling back to pure python.
try:
    from google_crc32c import cext as impl
    implementation = "c"
except ImportError as exc:
    from google_crc32c import python as impl  # type: ignore
    warnings.warn(_SLOW_CRC32C_WARNING, RuntimeWarning)
    implementation = "python"

extend = impl.extend
value = impl.value

Checksum = impl.Checksum

__all__ = ["extend", "value", "Checksum", "implementation"]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct


class CommonChecksum(object):
    """Hashlib-alike helper for CRC32C operations.

    This class should not be used directly and requires an update implementation.

    Args:
        initial_value (Optional[bytes]): the initial chunk of data from
            which the CRC32C checksum is computed.  Defaults to b''.
    """
    __slots__ = ()

    def __init__(self, initial_value=b""):
        self._crc = 0
        if initial_value != b"":
            self.update(initial_value)

    def update(self, data):
        """Update the checksum with a new chunk of data.

        Args:
            chunk (Optional[bytes]): a chunk of data used to extend
                the CRC32C checksum.
        """
        raise NotImplementedError

    def digest(self):
        """Big-endian order, per RFC 4960.

        See: https://cloud.google.com/storage/docs/json_api/v1/objects#crc32c

        Returns:
            bytes: An eight-byte digest string.
        """
        return struct.pack(">L", self._crc)

    def hexdigest(self):
        """Like :meth:`digest` except returns as a bytestring of double length.

        Returns
            bytes: A sixteen byte digest string, contaiing only hex digits.
        """
        return "{:08x}".format(self._crc).encode("ascii")

    def copy(self):
        """Create another checksum with the same CRC32C value.

        Returns:
            Checksum: the new instance.
        """
        clone = self.__class__()
        clone._crc = self._crc
        return clone

    def consume(self, stream, chunksize):
        """Consume chunks from a stream, extending our CRC32 checksum.

        Args:
            stream (BinaryIO): the stream to consume.
            chunksize (int): the size of the read to perform

        Returns:
            Generator[bytes, None, None]: Iterable of the chunks read from the
            stream.
        """
        while True:
            chunk = stream.read(chunksize)
            if not chunk:
                break
            self.update(chunk)
            yield chunk



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import shutil
import setuptools
import setuptools.command.build_ext

_EXTRA_DLL = "extra-dll"
_DLL_FILENAME = "crc32c.dll"

# Explicit environment variable disables pure-Python fallback
CRC32C_PURE_PYTHON_EXPLICIT = "CRC32C_PURE_PYTHON" in os.environ
_FALSE_OPTIONS = ("0", "false", "no", "False", "No", None)
CRC32C_PURE_PYTHON = os.getenv("CRC32C_PURE_PYTHON") not in _FALSE_OPTIONS


def copy_dll(build_lib):
    install_prefix = os.environ.get("CRC32C_INSTALL_PREFIX")

    if os.name == "nt" and install_prefix is not None:
        assert os.path.isdir(install_prefix)

        installed_dll = os.path.join(install_prefix, "bin", _DLL_FILENAME)
        assert os.path.isfile(installed_dll)

        lib_dlls = os.path.join(build_lib, "google_crc32c", _EXTRA_DLL)
        os.makedirs(lib_dlls, exist_ok=True)
        relocated_dll = os.path.join(lib_dlls, _DLL_FILENAME)

        shutil.copyfile(installed_dll, relocated_dll)
        assert os.path.isfile(relocated_dll)


class BuildExtWithDLL(setuptools.command.build_ext.build_ext):
    def run(self):
        copy_dll(self.build_lib)
        result = setuptools.command.build_ext.build_ext.run(self)
        return result


def build_pure_python():
    setuptools.setup(
        packages=["google_crc32c"],
        package_dir={"": "src"},
        ext_modules=[],
    )


def build_c_extension():
    install_prefix = os.getenv("CRC32C_INSTALL_PREFIX")
    if install_prefix is not None:
        install_prefix = os.path.normcase(install_prefix)
        print(f"#### using local install of 'crc32c': {install_prefix!r}")
        #assert os.path.isdir(install_prefix)
        install_prefix = os.path.realpath(install_prefix)
        include_dirs = [os.path.join(install_prefix, "include")]
        library_dirs = [os.path.join(install_prefix, "lib")]

        if platform.system() == "Linux":
            library_dirs.append(os.path.join(install_prefix, "lib64"))

        if os.name == "nt":
            library_dirs.append(os.path.join(install_prefix, "bin"))
            kwargs = {
                "include_dirs": include_dirs,
                "library_dirs": library_dirs,
            }
        else:
            runtime_library_dirs = library_dirs[:]
            kwargs = {
                "include_dirs": include_dirs,
                "library_dirs": library_dirs,
                "runtime_library_dirs": runtime_library_dirs,
            }
    else:
        print("#### using global install of 'crc32c'")
        kwargs = {}

    module_path = os.path.join("src", "google_crc32c", "_crc32c.c")
    sources=[os.path.normcase(module_path)]
    print(f"##### sources: {sources}")
    print(f"##### module kwargs: {kwargs}")
    module = setuptools.Extension(
        "google_crc32c._crc32c",
        sources=sources,
        libraries=["crc32c"],
        **kwargs
    )

    setuptools.setup(
        packages=["google_crc32c"],
        package_dir={"": "src"},
        ext_modules=[module],
        cmdclass={"build_ext": BuildExtWithDLL},
        install_requires=["importlib_resources>=1.3 ; python_version < '3.9' and os_name == 'nt'"],
    )


if CRC32C_PURE_PYTHON:
    build_pure_python()
else:
    try:
        build_c_extension()
    except SystemExit:
        if CRC32C_PURE_PYTHON_EXPLICIT:
            # If build / install fails, it is likely a compilation error with
            # the C extension:  advise user how to enable the pure-Python
            # build.
            logging.error(
                "Compiling the C Extension for the crc32c library failed. "
                "To enable building / installing a pure-Python-only version, "
                "set 'CRC32C_PURE_PYTHON=1' in the environment."
            )
            raise

        logging.info(
            "Compiling the C Extension for the crc32c library failed. "
            "Falling back to pure Python build."
        )
        build_pure_python()
